# Copyright 2021 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import json
import logging
import zlib

from collections.abc import MutableMapping
from urllib.parse import quote_plus, unquote_plus

from kazoo.exceptions import NoNodeError
from kazoo.recipe import lock

from zuul import model
from zuul.zk import sharding, ZooKeeperSimpleBase

CONFIG_ROOT = "/zuul/config"


def _safe_path(root_path, *keys):
    return "/".join((root_path, *(quote_plus(k) for k in keys)))


class FilesCache(ZooKeeperSimpleBase, MutableMapping):
    """Cache of raw config files in Zookeeper for a project-branch.

    Data will be stored in Zookeeper using the following path:
        /zuul/config/<project>/<branch>/<filename>

    """
    log = logging.getLogger("zuul.zk.config_cache.FilesCache")

    def __init__(self, client, root_path):
        super().__init__(client)
        self.root_path = root_path
        self.files_path = f"{root_path}/files"

    def setValidFor(self, extra_config_files, extra_config_dirs, ltime):
        """Set the cache valid for the given extra config files/dirs."""
        data = {
            "extra_files_searched": list(extra_config_files),
            "extra_dirs_searched": list(extra_config_dirs),
            "ltime": ltime,
        }
        payload = json.dumps(data, sort_keys=True).encode("utf8")
        try:
            self.kazoo_client.set(self.root_path, payload)
        except NoNodeError:
            self.kazoo_client.create(self.root_path, payload, makepath=True)

    def isValidFor(self, tpc, min_ltime):
        """Check if the cache is valid.

        Check if the cache is valid for the given tenant project config
        (tpc) and that it is up-to-date, relative to the give logical
        timestamp.

        """
        try:
            data, _ = self.kazoo_client.get(self.root_path)
        except NoNodeError:
            return False

        try:
            content = json.loads(data)
            extra_files_searched = set(content["extra_files_searched"])
            extra_dirs_searched = set(content["extra_dirs_searched"])
            ltime = content["ltime"]
        except Exception:
            return False

        if ltime < min_ltime:
            # Cache is outdated
            return False

        return (set(tpc.extra_config_files) <= extra_files_searched
                and set(tpc.extra_config_dirs) <= extra_dirs_searched)

    @property
    def ltime(self):
        try:
            data, _ = self.kazoo_client.get(self.root_path)
            content = json.loads(data)
            return content["ltime"]
        except Exception:
            return -1

    def _key_path(self, key):
        return _safe_path(self.files_path, key)

    def __getitem__(self, key):
        try:
            with sharding.BufferedShardReader(
                self.kazoo_client, self._key_path(key)
            ) as stream:
                return zlib.decompress(stream.read()).decode("utf8")
        except NoNodeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        path = self._key_path(key)
        with sharding.BufferedShardWriter(self.kazoo_client, path) as stream:
            stream.truncate(0)
            stream.write(zlib.compress(value.encode("utf8")))
            stream.flush()

    def __delitem__(self, key):
        try:
            self.kazoo_client.delete(self._key_path(key), recursive=True)
        except NoNodeError:
            raise KeyError(key)

    def __iter__(self):
        try:
            children = self.kazoo_client.get_children(self.files_path)
        except NoNodeError:
            children = []
        yield from sorted(unquote_plus(c) for c in children)

    def __len__(self):
        try:
            children = self.kazoo_client.get_children(self.files_path)
        except NoNodeError:
            children = []
        return len(children)

    def clear(self):
        with contextlib.suppress(NoNodeError):
            self.kazoo_client.delete(self.root_path, recursive=True)


class UnparsedConfigCache(ZooKeeperSimpleBase):
    """Zookeeper cache for unparsed config files."""

    log = logging.getLogger("zuul.zk.config_cache.UnparsedConfigCache")

    def __init__(self, client, config_root=CONFIG_ROOT):
        super().__init__(client)
        self.cache_path = f"{config_root}/cache"
        self.lock_path = f"{config_root}/lock"

    def readLock(self, project_cname):
        return lock.ReadLock(
            self.kazoo_client, _safe_path(self.lock_path, project_cname))

    def writeLock(self, project_cname):
        return lock.WriteLock(
            self.kazoo_client, _safe_path(self.lock_path, project_cname))

    def getFilesCache(self, project_cname, branch_name):
        path = _safe_path(self.cache_path, project_cname, branch_name)
        return FilesCache(self.client, path)

    def listCachedProjects(self):
        try:
            children = self.kazoo_client.get_children(self.cache_path)
        except NoNodeError:
            children = []
        yield from sorted(unquote_plus(c) for c in children)

    def clearCache(self, project_cname, branch_name=None):
        if branch_name is None:
            path = _safe_path(self.cache_path, project_cname)
        else:
            path = _safe_path(self.cache_path, project_cname, branch_name)
        with contextlib.suppress(NoNodeError):
            self.kazoo_client.delete(path, recursive=True)


class SystemConfigCache(ZooKeeperSimpleBase):
    """Zookeeper cache for Zuul system configuration.

    The system configuration consists of the unparsed abide config and
    the runtime related settings from the Zuul config file.
    """

    SYSTEM_ROOT = "/zuul/system"
    log = logging.getLogger("zuul.zk.SystemConfigCache")

    def __init__(self, client, callback):
        super().__init__(client)
        self.conf_path = f"{self.SYSTEM_ROOT}/conf"
        self.lock_path = f"{self.SYSTEM_ROOT}/conf-lock"
        self._callback = callback
        self.kazoo_client.DataWatch(self.conf_path, self._configChanged)

    def _configChanged(self, data, stat, event):
        self._callback()

    @property
    def ltime(self):
        with lock.ReadLock(self.kazoo_client, self.lock_path):
            zstat = self.kazoo_client.exists(self.conf_path)
        return -1 if zstat is None else zstat.last_modified_transaction_id

    @property
    def is_valid(self):
        return self.ltime > -1

    def get(self):
        """Get the system configuration from Zookeeper.

        :returns: A tuple (unparsed abide config, system attributes)
        """
        with lock.ReadLock(self.kazoo_client, self.lock_path):
            try:
                with sharding.BufferedShardReader(
                    self.kazoo_client, self.conf_path
                ) as stream:
                    data = json.loads(zlib.decompress(stream.read()))
                    zstat = stream.zstat
            except Exception:
                raise RuntimeError("No valid system config")
            return (model.UnparsedAbideConfig.fromDict(
                    data["unparsed_abide"],
                    ltime=zstat.last_modified_transaction_id
                ), model.SystemAttributes.fromDict(data["system_attributes"]))

    def set(self, unparsed_abide, system_attributes):
        with lock.WriteLock(self.kazoo_client, self.lock_path):
            data = {
                "unparsed_abide": unparsed_abide.toDict(),
                "system_attributes": system_attributes.toDict(),
            }
            with sharding.BufferedShardWriter(
                self.kazoo_client, self.conf_path
            ) as stream:
                stream.truncate(0)
                stream.write(zlib.compress(
                    json.dumps(data, sort_keys=True).encode("utf8")))
                stream.flush()
                zstat = stream.zstat
            unparsed_abide.ltime = zstat.last_modified_transaction_id



# Algorithm from http://www.logarithmic.net/pfh/blog/01208083168
# License: public domain
# Authors: Dr. Paul Harrison / Dries Verdegem


def strongly_connected_components(graph):
    """
    Tarjan's Algorithm (named for its discoverer, Robert Tarjan) is a graph
    theory algorithm for finding the strongly connected components of a graph.

    Based on:
    http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
    """

    index_counter = [0]
    stack = []
    lowlinks = {}
    index = {}
    result = []

    def strongconnect(node):
        # set the depth index for this node to the smallest unused index
        index[node] = index_counter[0]
        lowlinks[node] = index_counter[0]
        index_counter[0] += 1
        stack.append(node)

        # Consider successors of `node`
        try:
            successors = graph[node]
        except KeyError:
            successors = []
        for successor in successors:
            if successor not in lowlinks:
                # Successor has not yet been visited; recurse on it
                strongconnect(successor)
                lowlinks[node] = min(lowlinks[node], lowlinks[successor])
            elif successor in stack:
                # the successor is in the stack and hence in the current
                # strongly connected component (SCC)
                lowlinks[node] = min(lowlinks[node], index[successor])

        # If 'node' is a root node, pop the stack and generate an SCC
        if lowlinks[node] == index[node]:
            connected_component = []

            while True:
                successor = stack.pop()
                connected_component.append(successor)
                if successor == node:
                    break
            component = tuple(connected_component)
            # storing the result
            result.append(component)

    for node in graph:
        if node not in lowlinks:
            strongconnect(node)

    return result



# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import tempfile
import time
from unittest import mock

from zuul.driver.azure.azuremodel import AzureProviderNode
import zuul.driver.azure.azureendpoint

from tests import fake_azure
from tests.base import (
    ResponsesFixture,
    ZuulTestCase,
    driver_config,
    iterate_timeout,
    return_data,
    simple_layout,
)
from tests.unit.test_launcher import ImageMocksFixture
from tests.unit.test_cloud_driver import BaseCloudDriverTest


def make_image(name, tags):
    return {
        'name': name,
        'id': ('/subscriptions/c35cf7df-ed75-4c85-be00-535409a85120/'
               'resourceGroups/nodepool/providers/Microsoft.Compute/'
               f'images/{name}'),
        'type': 'Microsoft.Compute/images',
        'location': 'eastus',
        'tags': tags,
        'properties': {
            'storageProfile': {
                'osDisk': {
                    'osType': 'Linux',
                    'osState': 'Generalized',
                    'diskSizeGB': 1,
                    'blobUri': 'https://example.net/nodepoolstorage/img.vhd',
                    'caching': 'ReadWrite',
                    'storageAccountType': 'Standard_LRS'
                },
                'dataDisks': [],
                'zoneResilient': False
            },
            'provisioningState': 'Succeeded',
            'hyperVGeneration': 'V1'
        }
    }


class BaseAzureDriverTest(ZuulTestCase):
    cloud_test_image_format = 'vhd'
    cloud_test_provider_name = 'azure-main'
    config_file = 'zuul-connections-nodepool.conf'
    debian_return_data = {
        'zuul': {
            'artifacts': [
                {
                    'name': 'vhd image',
                    'url': 'http://example.com/image.vhd',
                    'metadata': {
                        'type': 'zuul_image',
                        'image_name': 'debian-local',
                        'format': 'vhd',
                        'sha256': ImageMocksFixture.vhd_sha256,
                        'md5sum': ImageMocksFixture.vhd_md5sum,
                    }
                },
            ]
        }
    }

    default_quotas = {
        'lowPriorityCores': 3,
        'cores': 4,
        'virtualMachines': 25000,
    }

    def setUp(self):
        self.initTestConfig()
        self.patch(zuul.driver.azure.azureendpoint, 'CACHE_TTL', 1)
        responses_fixture = self.useFixture(ResponsesFixture())
        self.useFixture(ImageMocksFixture(responses_fixture))

        quotas = self.default_quotas.copy()
        quotas.update(self.test_config.driver.azure.get('quotas', {}))

        self.fake_azure = fake_azure.FakeAzureFixture(
            responses_fixture, quotas)
        self.useFixture(self.fake_azure)

        super().setUp()

    @contextlib.contextmanager
    def _block_futures(self):
        with (mock.patch(
                'zuul.driver.azure.azureendpoint.'
                'AzureProviderEndpoint._completeApi', return_value=None)):
            yield


class TestAzureDriver(BaseAzureDriverTest, BaseCloudDriverTest):
    def _assertProviderNodeAttributes(self, pnode):
        super()._assertProviderNodeAttributes(pnode)
        self.assertEqual('Azure', pnode.cloud)
        self.assertEqual('centralus', pnode.region)
        self.assertEqual('1', pnode.az)
        if checks := self.test_config.driver.azure.get('node_checks'):
            checks(self, pnode)

    @simple_layout('layouts/azure/nodepool.yaml', enable_nodepool=True)
    def test_azure_node_lifecycle(self):
        self._test_node_lifecycle('debian-normal')

    def check_more_attrs(self, pnode):
        pass

    @simple_layout('layouts/azure/more.yaml', enable_nodepool=True)
    @driver_config('azure', node_checks=check_more_attrs)
    def test_azure_node_lifecycle_more(self):
        # Test with more options than the normal test
        # windows-generate
        # ephemeral disk
        self._test_node_lifecycle('debian-normal')
        vm_request = (
            self.fake_azure.crud['Microsoft.Compute/virtualMachines'].
            requests[0])
        osDisk = vm_request['properties']['storageProfile']['osDisk']
        self.assertIn('caching', osDisk)
        self.assertEqual(osDisk['caching'], 'ReadOnly')
        self.assertIn('diffDiskSettings', osDisk)
        self.assertIn('option', osDisk['diffDiskSettings'])
        self.assertEqual(osDisk['diffDiskSettings']['option'], 'Local')

        self.assertEqual(
            vm_request['properties']['osProfile']['adminUsername'],
            'foobar')
        self.assertEqual(
            len(vm_request['properties']['osProfile']['adminPassword']),
            64)

    @simple_layout('layouts/azure/nodepool.yaml', enable_nodepool=True)
    @driver_config('azure', quotas={'cores': 1})
    def test_azure_quota(self):
        self._test_quota('debian-normal')

    @simple_layout('layouts/azure/nodepool-spot.yaml', enable_nodepool=True)
    @driver_config('azure', quotas={'lowPriorityCores': 1})
    def test_azure_quota_spot(self):
        self._test_quota('debian-normal')

    @simple_layout('layouts/azure/resource-limits.yaml',
                   enable_nodepool=True)
    def test_azure_resource_limits(self):
        self._test_quota('debian-normal')

    @simple_layout('layouts/azure/nodepool-image.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        BaseAzureDriverTest.debian_return_data,
    )
    def test_azure_diskimage(self):
        self._test_diskimage()

    @simple_layout('layouts/azure/nodepool.yaml', enable_nodepool=True)
    def test_state_machines(self):
        label_name = "debian-normal"
        provider_name = "azure-main"
        node_class = AzureProviderNode
        future_names = ['delete_future', 'create_future']
        self._test_state_machines(label_name, provider_name,
                                  node_class, future_names)

    # Azure-driver specific tests

    @simple_layout('layouts/azure/nodepool.yaml',
                   enable_nodepool=True)
    def test_azure_resource_cleanup(self):
        self.waitUntilSettled()
        self.launcher.cleanup_worker.INTERVAL = 1
        provider = self.launcher._getProvider(
            'tenant-one', 'azure-main')
        endpoint = provider.getEndpoint()
        client = endpoint._client

        system_id = self.launcher.system.system_id
        tags = {
            'zuul_system_id': system_id,
            'zuul_node_uuid': '0000000042',
        }

        subnet_id = ('/subscriptions/c35cf7df-ed75-4c85-be00-535409a85120/'
                     'resourceGroups/nodepool/providers/Microsoft.Network/'
                     'virtualNetworks/NodePool/subnets/default')
        pip_spec = {
            'name': 'npf2d7ff7d081c4-v4-ip',
            'tags': tags,
            'properties': {
                'deleteOption': 'Delete',
                'publicIPAllocationMethod': 'Static',
                'publicIPAddressVersion': 'IPv4',
            },
            'sku': {
                'name': 'Standard',
            },
        }
        vm_spec = {
            'location': 'centralus',
            'properties': {
                'hardwareProfile': {'vmSize': 'Standard_B1ls'},
                'networkProfile': {
                    'networkApiVersion': '2020-11-01',
                    'networkInterfaceConfigurations': [
                        {'name': 'npf2d7ff7d081c4-nic',
                         'tags': tags,
                         'properties': {
                             'deleteOption': 'Delete',
                             'ipConfigurations': [
                                 {'name': 'npf2d7ff7d081c4-v4-config',
                                  'properties': {
                                      'privateIPAddressVersion': 'IPv4',
                                      'subnet': {
                                          'id': subnet_id,
                                      },
                                      'publicIPAddressConfiguration': pip_spec,
                                  }}]}}]},
                'osProfile': {
                    'computerName': 'npf2d7ff7d081c4',
                },
                'storageProfile': {
                    'imageReference': {'offer': 'UbuntuServer',
                                       'publisher': 'Canonical',
                                       'sku': '18.04-LTS',
                                       'version': 'latest'},
                    'osDisk': {'createOption': 'FromImage',
                               'deleteOption': 'Delete'}}},
            'tags': tags,
        }

        client.resource_group_virtual_machines.create(
            'otherrg', 'npf2d7ff7d081c4', vm_spec)

        image_spec = {
            'location': 'centralus',
            'tags': tags,
            'properties': {
                'hyperVGeneration': 'V2',
                'storageProfile': {
                    'osDisk': {
                        'osType': 'Linux',
                        'managedDisk': {
                            'id': 'testdisk',
                        },
                        'osState': 'Generalized'
                    },
                    'zoneResilient': True
                }
            }
        }
        image = client.resource_group_images.create(
            'otherrg', 'testimage', image_spec)
        image = client.wait_for_async_operation(image)

        self.assertEqual(1, len(client.virtual_machines.list()))
        self.assertEqual(1, len(client.network_interfaces.list()))
        self.assertEqual(1, len(client.public_ip_addresses.list()))
        self.assertEqual(1, len(client.disks.list()))
        self.assertEqual(1, len(client.images.list()))

        self.log.debug("Start cleanup worker")
        self.launcher.cleanup_worker.start()

        for _ in iterate_timeout(30, 'vm deletion'):
            if not client.virtual_machines.list():
                break
            time.sleep(1)
        for _ in iterate_timeout(30, 'nic deletion'):
            if not client.network_interfaces.list():
                break
            time.sleep(1)
        for _ in iterate_timeout(30, 'pip deletion'):
            if not client.public_ip_addresses.list():
                break
            time.sleep(1)
        for _ in iterate_timeout(30, 'disk deletion'):
            if not client.disks.list():
                break
            time.sleep(1)
        for _ in iterate_timeout(30, 'image deletion'):
            if not client.images.list():
                break
            time.sleep(1)

    @simple_layout('layouts/azure/nodepool-images.yaml', enable_nodepool=True)
    def test_azure_image_id(self):
        self._test_node_lifecycle('debian-normal-id')
        self.assertEqual(
            self.fake_azure.crud['Microsoft.Compute/virtualMachines'].
            requests[0]['properties']['storageProfile']
            ['imageReference']['id'],
            "/subscriptions/c35cf7df-ed75-4c85-be00-535409a85120"
            "/resourceGroups/nodepool/providers/Microsoft.Compute"
            "/images/test-image-1234")

    @simple_layout('layouts/azure/nodepool-images.yaml', enable_nodepool=True)
    def test_azure_image_filter(self):
        self.fake_azure.crud['Microsoft.Compute/images'].items.append(
            make_image('test1', {'foo': 'bar'}))
        self._test_node_lifecycle('debian-normal-filter')
        self.assertEqual(
            self.fake_azure.crud['Microsoft.Compute/virtualMachines'].
            requests[0]['properties']['storageProfile']
            ['imageReference']['id'],
            "/subscriptions/c35cf7df-ed75-4c85-be00-535409a85120"
            "/resourceGroups/nodepool/providers/Microsoft.Compute"
            "/images/test1")

    @simple_layout('layouts/azure/nodepool-images.yaml', enable_nodepool=True)
    def test_azure_image_community_gallery(self):
        self._test_node_lifecycle('debian-normal-community')
        self.assertEqual(
            self.fake_azure.crud['Microsoft.Compute/virtualMachines'].
            requests[0]['properties']['storageProfile']
            ['imageReference']['communityGalleryImageId'],
            "/CommunityGalleries/community-gallery"
            "/Images/community-image"
            "/Versions/latest")

    @simple_layout('layouts/azure/nodepool-images.yaml', enable_nodepool=True)
    def test_azure_image_shared_gallery(self):
        self._test_node_lifecycle('debian-normal-shared')
        self.assertEqual(
            self.fake_azure.crud['Microsoft.Compute/virtualMachines'].
            requests[0]['properties']['storageProfile']
            ['imageReference']['sharedGalleryImageId'],
            "/SharedGalleries/shared-gallery"
            "/Images/shared-image"
            "/Versions/latest")


class TestAzureDriverOidc(BaseAzureDriverTest, BaseCloudDriverTest):
    config_file = 'zuul-connections-azure-oidc.conf'

    def setUp(self):
        self.token_file = tempfile.NamedTemporaryFile('w', delete=False)
        with self.token_file as f:
            f.write("testtoken")
        super().setUp()

    def setup_config(self, config_file):
        config = super().setup_config(config_file)
        config.set('connection azure', 'federated_token_file',
                   self.token_file.name)
        return config

    @simple_layout('layouts/azure/nodepool.yaml', enable_nodepool=True)
    def test_azure_node_lifecycle(self):
        self._test_node_lifecycle('debian-normal')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import os
import time
from enum import Enum
from typing import Optional
import threading

from kazoo.exceptions import NoNodeError, LockTimeout
from kazoo.recipe.cache import TreeCache, TreeEvent
from kazoo.recipe.lock import Lock

import zuul.model
from zuul.lib import tracing
from zuul.lib.jsonutil import json_dumps
from zuul.model import HoldRequest, NodeRequest, Node
from zuul.zk import ZooKeeperBase
from zuul.zk.exceptions import LockException


class NodeRequestEvent(Enum):
    COMPLETED = 0  # FULFILLED | FAILED
    DELETED = 2


class NodeEvent(Enum):
    CHANGED = 0
    DELETED = 2


class ZooKeeperNodepool(ZooKeeperBase):
    """
    Class implementing Nodepool related ZooKeeper interface.
    """
    NODES_ROOT = "/nodepool/nodes"
    COMPONENT_ROOT = "/nodepool/components"
    REQUEST_ROOT = '/nodepool/requests'
    REQUEST_LOCK_ROOT = "/nodepool/requests-lock"
    HOLD_REQUEST_ROOT = '/zuul/hold-requests'

    log = logging.getLogger("zuul.zk.nodepool.ZooKeeperNodepool")

    def __init__(self, client,
                 enable_node_request_cache=False,
                 node_request_event_callback=None,
                 connection_suspended_callback=None,
                 enable_node_cache=False):
        super().__init__(client)
        self.enable_node_request_cache = enable_node_request_cache
        self.node_request_event_callback = node_request_event_callback
        self.connection_suspended_callback = connection_suspended_callback
        self.enable_node_cache = enable_node_cache
        # The caching model we use is designed around handing out model
        # data as objects. To do this, we use two caches: one is a TreeCache
        # which contains raw znode data (among other details), and one for
        # storing that data serialized as objects. This allows us to return
        # objects from the APIs, and avoids calling the methods to serialize
        # the data into objects more than once.
        self._node_request_tree = None
        self._node_request_cache = {}
        self._node_tree = None
        self._node_cache = {}
        self._callback_lock = threading.Lock()

        if self.client.connected:
            self._onConnect()

    def _onConnect(self):
        if self.enable_node_request_cache:
            self._node_request_tree = TreeCache(self.kazoo_client,
                                                self.REQUEST_ROOT)
            self._node_request_tree.listen_fault(self._cacheFaultListener)
            self._node_request_tree.listen(self.requestCacheListener)
            self._node_request_tree.start()
        if self.enable_node_cache:
            self._node_tree = TreeCache(self.kazoo_client,
                                        self.NODES_ROOT)
            self._node_tree.listen_fault(self._cacheFaultListener)
            self._node_tree.listen(self.nodeCacheListener)
            self._node_tree.start()

    def _onDisconnect(self):
        if self._node_request_tree is not None:
            self._node_request_tree.close()
            self._node_request_tree = None
        if self._node_tree is not None:
            self._node_tree.close()
            self._node_tree = None

    def _onSuspended(self):
        if self.connection_suspended_callback:
            self.connection_suspended_callback()

    def _nodePath(self, node):
        return "%s/%s" % (self.NODES_ROOT, node)

    def _cacheFaultListener(self, e):
        self.log.exception(e)

    def register(self):
        self.kazoo_client.ensure_path(self.REQUEST_ROOT)

    def getRegisteredLaunchers(self):
        """
        Get a list of all launchers that have registered with ZooKeeper.

        :returns: A list of Launcher objects, or empty list if none are found.
        """
        root_path = os.path.join(self.COMPONENT_ROOT, 'pool')
        try:
            pools = self.kazoo_client.get_children(root_path)
        except NoNodeError:
            return []

        objs = []
        for pool in pools:
            path = os.path.join(root_path, pool)
            try:
                data, _ = self.kazoo_client.get(path)
            except NoNodeError:
                # launcher disappeared
                continue

            objs.append(Launcher.fromDict(json.loads(data.decode('utf8'))))
        return objs

    def getNodes(self, cached=False):
        """
        Get the current list of all node ids.

        :param bool cached: Whether to use the internal cache to get the list
            of ids.
        :returns: A list of node ids.
        """
        if cached and self.enable_node_cache:
            return list(self._node_cache.keys())
        else:
            try:
                return self.kazoo_client.get_children(self.NODES_ROOT)
            except NoNodeError:
                return []

    def _getNodeData(self, node):
        """
        Get the data for a specific node.

        :param str node: The node ID.

        :returns: The node data, or None if the node was not found.
        """
        path = self._nodePath(node)
        try:
            data, stat = self.kazoo_client.get(path)
        except NoNodeError:
            return None
        if not data:
            return None

        d = json.loads(data.decode('utf8'))
        d['id'] = node
        return d

    def getNode(self, node_id):
        """
        Get a Node object for a specific node.

        :param str node_id: The node ID.

        :returns: The Node, or None if the node was not found.
        """
        node = self._node_cache.get(node_id)
        if node:
            return node

        path = self._nodePath(node_id)
        try:
            data, stat = self.kazoo_client.get(path)
        except NoNodeError:
            return None
        if not data:
            return None

        node = Node(None, None)
        node.updateFromDict(data)
        node.id = node_id
        node.stat = stat
        return node

    def nodeIterator(self, cached=False):
        """
        Utility generator method for iterating through all nodes.
        """
        for node_id in self.getNodes(cached):
            node = self.getNode(node_id)
            if node:
                yield node

    def getHoldRequests(self):
        """
        Get the current list of all hold requests.
        """

        try:
            return sorted(self.kazoo_client
                          .get_children(self.HOLD_REQUEST_ROOT))
        except NoNodeError:
            return []

    def getHoldRequest(self, hold_request_id):
        # To be friendly, zero pad if this request came from a client
        # call, where the user might have done "zuul autohold-delete
        # 123" thinking that would match the "0000000123" shown in
        # "autohold-list".
        #
        #  "A sequential node will be given the specified path plus a
        #  suffix i where i is the current sequential number of the
        #  node. The sequence number is always fixed length of 10
        #  digits, 0 padded. Once such a node is created, the
        #  sequential number will be incremented by one."
        if len(hold_request_id) != 10:
            hold_request_id = hold_request_id.rjust(10, '0')

        path = self.HOLD_REQUEST_ROOT + "/" + hold_request_id
        try:
            data, stat = self.kazoo_client.get(path)
        except NoNodeError:
            return None
        if not data:
            return None

        obj = HoldRequest.fromDict(json.loads(data.decode('utf8')))
        obj.id = hold_request_id
        obj.stat = stat
        return obj

    def storeHoldRequest(self, request: HoldRequest):
        """
        Create or update a hold request.

        If this is a new request with no value for the `id` attribute of the
        passed in request, then `id` will be set with the unique request
        identifier after successful creation.

        :param HoldRequest request: Object representing the hold request.
        """
        if request.id is None:
            path = self.kazoo_client.create(
                self.HOLD_REQUEST_ROOT + "/",
                value=request.serialize(),
                sequence=True,
                makepath=True)
            request.id = path.split('/')[-1]
        else:
            path = self.HOLD_REQUEST_ROOT + "/" + request.id
            self.kazoo_client.set(path, request.serialize())

    def _markHeldNodesAsUsed(self, hold_request, node_group):
        """
        Changes the state for each held node for the hold request to 'used'.

        :returns: True if all nodes marked USED, False otherwise.
        """

        failure = False
        for node_id in node_group['nodes']:
            node = self._getNodeData(node_id)
            if not node or node['state'] == zuul.model.STATE_USED:
                continue

            node['state'] = zuul.model.STATE_USED

            name = None
            label = None
            if 'name' in node:
                name = node['name']
            if 'label' in node:
                label = node['label']

            node_obj = zuul.model.Node(name, label)
            node_obj.updateFromDict(node)

            try:
                self.lockNode(node_obj, blocking=False)
                self.storeNode(node_obj)
            except Exception:
                self.log.exception("Cannot change HELD node state to USED "
                                   "for node %s in request %s",
                                   node_obj.id, hold_request.id)
                failure = True
            finally:
                try:
                    if node_obj.lock:
                        self.unlockNode(node_obj)
                except Exception:
                    self.log.exception(
                        "Failed to unlock HELD node %s for request %s",
                        node_obj.id, hold_request.id)

        return not failure

    def deleteHoldRequest(self, hold_request, launcher_client):
        """
        Delete a hold request.

        :param HoldRequest request: Object representing the hold request.
        """
        for node_group in hold_request.nodes:
            if node_group.get('niz', False):
                if not launcher_client.markHeldNodesAsUsed(
                        hold_request, node_group):
                    self.log.info("Unable to delete hold request %s because "
                                  "not all nodes marked as USED.",
                                  hold_request.id)
                    return
            else:
                if not self._markHeldNodesAsUsed(
                        hold_request, node_group):
                    self.log.info("Unable to delete hold request %s because "
                                  "not all nodes marked as USED.",
                                  hold_request.id)
                    return

        path = self.HOLD_REQUEST_ROOT + "/" + hold_request.id
        try:
            self.kazoo_client.delete(path, recursive=True)
        except NoNodeError:
            pass

    def lockHoldRequest(self, request: HoldRequest,
                        blocking: bool = True, timeout: Optional[int] = None):
        """
        Lock a node request.

        This will set the `lock` attribute of the request object when the
        lock is successfully acquired.

        :param request: The hold request to lock.
        :param blocking: Block until lock is obtained or return immediately.
        :param timeout: Don't wait forever to acquire the lock.
        """
        if not request.id:
            raise LockException(
                "Hold request without an ID cannot be locked: %s" % request)

        path = "%s/%s/lock" % (self.HOLD_REQUEST_ROOT, request.id)
        try:
            lock = Lock(self.kazoo_client, path)
            have_lock = lock.acquire(blocking, timeout)
        except LockTimeout:
            raise LockException("Timeout trying to acquire lock %s" % path)

        # If we aren't blocking, it's possible we didn't get the lock
        # because someone else has it.
        if not have_lock:
            raise LockException("Did not get lock on %s" % path)

        request.lock = lock

    def unlockHoldRequest(self, request: HoldRequest):
        """
        Unlock a hold request.

        The request must already have been locked.

        :param HoldRequest request: The request to unlock.

        :raises: ZKLockException if the request is not currently locked.
        """
        if request.lock is None:
            raise LockException("Request %s does not hold a lock" % request)
        request.lock.release()
        request.lock = None

    def requestCacheListener(self, event):
        try:
            # This lock is only needed for tests.  It wraps updating
            # our internal cache and performing the callback which
            # will generate node provisioned events in a critical
            # section.  This way the tests know if a request is
            # outstanding because either our cache is out of date or
            # there is an event on the queue.
            with self._callback_lock:
                self._requestCacheListener(event)
        except Exception:
            self.log.exception(
                "Exception in request cache update for event: %s",
                event)

    def _requestCacheListener(self, event):
        if hasattr(event.event_data, 'path'):
            # Ignore root node
            path = event.event_data.path
            if path == self.REQUEST_ROOT:
                return

            # Ignore lock nodes
            if '/lock' in path:
                return

        # Ignore any non-node related events such as connection events here
        if event.event_type not in (TreeEvent.NODE_ADDED,
                                    TreeEvent.NODE_UPDATED,
                                    TreeEvent.NODE_REMOVED):
            return

        path = event.event_data.path
        request_id = path.rsplit('/', 1)[1]

        if event.event_type in (TreeEvent.NODE_ADDED, TreeEvent.NODE_UPDATED):
            # Requests with empty data are invalid so skip add or update these.
            if not event.event_data.data:
                return

            # Perform an in-place update of the cached request if possible
            d = self._bytesToDict(event.event_data.data)
            request = self._node_request_cache.get(request_id)
            if request:
                if event.event_data.stat.version <= request.stat.version:
                    # Don't update to older data
                    return
                old_state = request.state
                request.updateFromDict(d)
                request.stat = event.event_data.stat
            else:
                old_state = None
                request = NodeRequest.fromDict(d)
                request.id = request_id
                request.stat = event.event_data.stat
                self._node_request_cache[request_id] = request

            if (request.state in {zuul.model.STATE_FULFILLED,
                                  zuul.model.STATE_FAILED}
                and request.state != old_state
                and self.node_request_event_callback):
                self.node_request_event_callback(
                    request, NodeRequestEvent.COMPLETED)

        elif event.event_type == TreeEvent.NODE_REMOVED:
            request = self._node_request_cache.pop(request_id)
            if self.node_request_event_callback:
                self.node_request_event_callback(
                    request, NodeRequestEvent.DELETED)

    def submitNodeRequest(self, node_request, priority):
        """
        Submit a request for nodes to Nodepool.

        :param NodeRequest node_request: A NodeRequest with the
            contents of the request.
        """
        node_request.created_time = time.time()
        node_request.span_info = tracing.startSavedSpan(
            "RequestNodes", start_time=node_request.created_time)
        data = node_request.toDict()

        path = '{}/{:0>3}-'.format(self.REQUEST_ROOT, priority)
        path = self.kazoo_client.create(path, json.dumps(data).encode('utf8'),
                                        makepath=True, sequence=True)
        reqid = path.split("/")[-1]
        node_request.id = reqid

    def getNodeRequests(self, cached=False):
        '''
        Get the current list of all node request IDs in priority sorted order.

        :param bool cached: Whether to use the internal cache to get the list
            of ids.
        :returns: A list of request ids.
        '''
        if cached and self.enable_node_request_cache:
            requests = self._node_request_cache.keys()
        else:
            try:
                requests = self.kazoo_client.get_children(self.REQUEST_ROOT)
            except NoNodeError:
                return []

        return sorted(requests)

    def getNodeRequest(self, node_request_id, cached=False):
        """
        Retrieve a NodeRequest from a given path in ZooKeeper.

        :param str node_request_id: The ID of the node request to retrieve.
        :param bool cached: Whether to use the cache.
        """
        if cached:
            req = self._node_request_cache.get(node_request_id)
            if req:
                return req

        path = f"{self.REQUEST_ROOT}/{node_request_id}"
        try:
            data, stat = self.kazoo_client.get(path)
        except NoNodeError:
            return None

        if not data:
            return None

        json_data = json.loads(data.decode("utf-8"))

        obj = NodeRequest.fromDict(json_data)
        obj.id = node_request_id
        obj.stat = stat
        return obj

    def deleteNodeRequest(self, node_request_id):
        """
        Delete a request for nodes.

        :param str node_request_id: The node request ID.

        :returns: True if the request was deleted, False if it did not exist.
        """
        path = '%s/%s' % (self.REQUEST_ROOT, node_request_id)
        try:
            self.kazoo_client.delete(path)
            return True
        except NoNodeError:
            return False

    def nodeRequestExists(self, node_request):
        """
        See if a NodeRequest exists in ZooKeeper.

        :param NodeRequest node_request: A NodeRequest to verify.

        :returns: True if the request exists, False otherwise.
        """
        path = '%s/%s' % (self.REQUEST_ROOT, node_request.id)
        if self.kazoo_client.exists(path):
            return True
        return False

    def storeNodeRequest(self, node_request):
        """
        Store the node request.

        The request is expected to already exist and is updated in its
        entirety.

        :param NodeRequest node_request: The request to update.
        """
        path = '%s/%s' % (self.REQUEST_ROOT, node_request.id)
        self.kazoo_client.set(
            path, json.dumps(node_request.toDict()).encode('utf8'),
            version=node_request.stat.version)

    def updateNodeRequest(self, node_request, data=None):
        """
        Refresh an existing node request.

        :param NodeRequest node_request: The request to update.
        :param dict data: The data to use; query ZK if absent.
        """
        if data is None:
            path = '%s/%s' % (self.REQUEST_ROOT, node_request.id)
            data, stat = self.kazoo_client.get(path)
        data = json.loads(data.decode('utf8'))
        node_request.updateFromDict(data)

    def nodeCacheListener(self, event):
        try:
            self._nodeCacheListener(event)
        except Exception:
            self.log.exception(
                "Exception in node cache update for event: %s",
                event)

    def _nodeCacheListener(self, event):
        # Ignore events without data
        if event.event_data is None:
            # This seems to be the case for some NODE_REMOVED events.
            # Without the data we are not able to pop() the correct node
            # from our cache dict. As we can't do much else, ignore the
            # event.
            return

        if hasattr(event.event_data, 'path'):
            # Ignore root node
            path = event.event_data.path
            if path == self.NODES_ROOT:
                return

            # Ignore lock nodes
            if '/lock' in path:
                return

        # Ignore any non-node related events such as connection events here
        if event.event_type not in (TreeEvent.NODE_ADDED,
                                    TreeEvent.NODE_UPDATED,
                                    TreeEvent.NODE_REMOVED):
            return

        path = event.event_data.path
        node_id = path.rsplit('/', 1)[1]

        if event.event_type in (TreeEvent.NODE_ADDED, TreeEvent.NODE_UPDATED):
            # Nodes with empty data are invalid so skip add or update these.
            if not event.event_data.data:
                return

            # Perform an in-place update of the cached node if possible
            d = self._bytesToDict(event.event_data.data)
            node = self._node_cache.get(node_id)
            if node:
                if event.event_data.stat.version <= node.stat.version:
                    # Don't update to older data
                    return
                node.updateFromDict(d)
                node.stat = event.event_data.stat
            else:
                node = Node(None, None)
                node.updateFromDict(d)
                node.id = node_id
                node.stat = event.event_data.stat
                self._node_cache[node_id] = node

        elif event.event_type == TreeEvent.NODE_REMOVED:
            node = self._node_cache.pop(node_id, None)

    def storeNode(self, node):
        """
        Store the node.

        The node is expected to already exist and is updated in its
        entirety.

        :param Node node: The node to update.
        """
        path = '%s/%s' % (self.NODES_ROOT, node.id)
        self.kazoo_client.set(path, json.dumps(node.toDict()).encode('utf8'))

    def updateNode(self, node, node_id):
        """
        Refresh an existing node.

        :param Node node: The node to update.

        :param str node_id: The ID of the node to update.  The
            existing node.id attribute (if any) will be ignored
            and replaced with this.
        """
        node_path = '%s/%s' % (self.NODES_ROOT, node_id)
        node.id = node_id
        node_data, node_stat = self.kazoo_client.get(node_path)
        node_data = json.loads(node_data.decode('utf8'))
        node.updateFromDict(node_data)

    def lockNode(self, node, blocking=True, timeout=None):
        """
        Lock a node.

        This should be called as soon as a request is fulfilled and
        the lock held for as long as the node is in-use.  It can be
        used by nodepool to detect if Zuul has gone offline and the
        node should be reclaimed.

        :param Node node: The node which should be locked.
        """
        lock_path = '%s/%s/lock' % (self.NODES_ROOT, node.id)
        try:
            lock = Lock(self.kazoo_client, lock_path)
            have_lock = lock.acquire(blocking, timeout)
        except LockTimeout:
            raise LockException(
                "Timeout trying to acquire lock %s" % lock_path)

        # If we aren't blocking, it's possible we didn't get the lock
        # because someone else has it.
        if not have_lock:
            raise LockException("Did not get lock on %s" % lock_path)

        node.lock = lock

    def unlockNode(self, node):
        """
        Unlock a node.

        The node must already have been locked.

        :param Node node: The node which should be unlocked.
        """

        if node.lock is None:
            raise LockException(f"Node {node} does not hold a lock")
        node.lock.release()
        node.lock = None

    def lockNodeRequest(self, request, blocking=True, timeout=None):
        """
        Lock a node request.

        This will set the `lock` attribute of the request object when the
        lock is successfully acquired.

        :param NodeRequest request: The request to lock.
        :param bool blocking: Whether or not to block on trying to
            acquire the lock
        :param int timeout: When blocking, how long to wait for the lock
            to get acquired. None, the default, waits forever.

        :raises: TimeoutException if we failed to acquire the lock when
            blocking with a timeout. ZKLockException if we are not blocking
            and could not get the lock, or a lock is already held.
        """
        path = "%s/%s" % (self.REQUEST_LOCK_ROOT, request.id)
        lock = Lock(self.kazoo_client, path)
        try:
            have_lock = lock.acquire(blocking, timeout)
        except LockTimeout:
            raise LockException(
                "Timeout trying to acquire lock %s" % path)
        except NoNodeError:
            have_lock = False
            self.log.error("Request not found for locking: %s", request)

        # If we aren't blocking, it's possible we didn't get the lock
        # because someone else has it.
        if not have_lock:
            raise LockException("Did not get lock on %s" % path)

        request.lock = lock

    def unlockNodeRequest(self, request):
        """
        Unlock a node request.

        The request must already have been locked.

        :param NodeRequest request: The request to unlock.

        :raises: ZKLockException if the request is not currently locked.
        """
        if request.lock is None:
            raise LockException(
                "Request %s does not hold a lock" % request)
        request.lock.release()
        request.lock = None

    def heldNodeCount(self, autohold_key):
        """
        Count the number of nodes being held for the given tenant/project/job.

        :param set autohold_key: A set with the tenant/project/job names.
        """
        identifier = " ".join(autohold_key)
        try:
            nodes = self.kazoo_client.get_children(self.NODES_ROOT)
        except NoNodeError:
            return 0

        count = 0
        for nodeid in nodes:
            node_path = '%s/%s' % (self.NODES_ROOT, nodeid)
            try:
                node_data, node_stat = self.kazoo_client.get(node_path)
            except NoNodeError:
                # Node got removed on us. Just ignore.
                continue

            if not node_data:
                self.log.warning("Node ID %s has no data", nodeid)
                continue
            node_data = json.loads(node_data.decode('utf8'))
            if (node_data['state'] == zuul.model.STATE_HOLD and
                    node_data.get('hold_job') == identifier):
                count += 1
        return count

    @staticmethod
    def _bytesToDict(data):
        return json.loads(data.decode("utf-8"))

    @staticmethod
    def _dictToBytes(data):
        # The custom json_dumps() will also serialize MappingProxyType objects
        return json_dumps(data).encode("utf-8")


class Launcher:
    """
    Class to describe a nodepool launcher.
    """

    def __init__(self):
        self.id = None
        self._supported_labels = set()

    def __eq__(self, other):
        if isinstance(other, Launcher):
            return (self.id == other.id and
                    self.supported_labels == other.supported_labels)
        else:
            return False

    @property
    def supported_labels(self):
        return self._supported_labels

    @supported_labels.setter
    def supported_labels(self, value):
        if not isinstance(value, set):
            raise TypeError("'supported_labels' attribute must be a set")
        self._supported_labels = value

    @staticmethod
    def fromDict(d):
        obj = Launcher()
        obj.id = d.get('id')
        obj.supported_labels = set(d.get('supported_labels', []))
        return obj



# Copyright 2024 BMW Group
# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul import model
from zuul.provider import statemachine


class SubnodeProviderNode(model.ProviderNode, subclass_id="subnode"):
    SUBNODE_ATTRIBUTES = (
        'uuid',
        'slot',
    )

    def __repr__(self):
        # Override the repr method to include the main node id
        return (f"<{self.__class__.__name__} uuid={self.uuid},"
                f" label={self.label}, state={self.state},"
                f" main_node={self.main_node_id}>")

    def updateFromMainNode(self, node):
        # Once the main node is finished creating, this is called to
        # synchronize any attributes that should be copied to the
        # subnode.
        # Called with an active context
        for k, v in node.getNodeData(serialize_node=True).items():
            if k in self.SUBNODE_ATTRIBUTES:
                continue
            setattr(self, k, v)


class SubnodeStateMachine(statemachine.StateMachine):
    """A no-op state machine since subnodes don't need to be created
    or deleted.

    """
    COMPLETE = 'complete'

    def __init__(self):
        super().__init__({})
        self.state = self.COMPLETE
        self.complete = True

    def advance(self):
        pass



# Copyright 2019 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re
import git
import yaml
import socket

from testtools.matchers import MatchesRegex

from zuul.lib import strings
from zuul.zk.layout import LayoutState

from tests.base import ZuulTestCase, simple_layout
from tests.base import ZuulWebFixture

EMPTY_LAYOUT_STATE = LayoutState("", "", 0, None, {}, -1)


class TestPagureDriver(ZuulTestCase):
    config_file = 'zuul-pagure-driver.conf'

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_pull_request_opened(self):

        initial_comment = "This is the\nPR initial_comment."
        A = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'A', initial_comment=initial_comment)
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)

        job = self.getJobFromHistory('project-test2')
        zuulvars = job.parameters['zuul']
        self.assertEqual(str(A.number), zuulvars['change'])
        self.assertEqual(str(A.commit_stop), zuulvars['patchset'])
        self.assertEqual(str(A.commit_stop), zuulvars['commit_id'])
        self.assertEqual('master', zuulvars['branch'])
        self.assertEqual('https://pagure/org/project/pull-request/1',
                         zuulvars['items'][0]['change_url'])
        self.assertEqual(zuulvars["message"],
                         strings.b64encode(initial_comment))
        self.assertEqual(2, len(self.history))
        self.assertEqual(2, len(A.comments))
        self.assertEqual(
            A.comments[0]['comment'], "Starting check jobs.")
        self.assertThat(
            A.comments[1]['comment'],
            MatchesRegex(r'.*\[project-test1 \]\(.*\).*', re.DOTALL))
        self.assertThat(
            A.comments[1]['comment'],
            MatchesRegex(r'.*\[project-test2 \]\(.*\).*', re.DOTALL))
        self.assertEqual(1, len(A.flags))
        self.assertEqual('success', A.flags[0]['status'])

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_pull_request_updated(self):

        A = self.fake_pagure.openFakePullRequest('org/project', 'master', 'A')
        pr_tip1 = A.commit_stop
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))
        self.assertHistory(
            [
                {'name': 'project-test1', 'changes': '1,%s' % pr_tip1},
                {'name': 'project-test2', 'changes': '1,%s' % pr_tip1},
            ], ordered=False
        )

        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        pr_tip2 = A.commit_stop
        self.waitUntilSettled()
        self.assertEqual(4, len(self.history))
        self.assertHistory(
            [
                {'name': 'project-test1', 'changes': '1,%s' % pr_tip1},
                {'name': 'project-test2', 'changes': '1,%s' % pr_tip1},
                {'name': 'project-test1', 'changes': '1,%s' % pr_tip2},
                {'name': 'project-test2', 'changes': '1,%s' % pr_tip2}
            ], ordered=False
        )

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_pull_request_updated_builds_aborted(self):

        A = self.fake_pagure.openFakePullRequest('org/project', 'master', 'A')
        pr_tip1 = A.commit_stop

        self.executor_server.hold_jobs_in_build = True

        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()

        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        pr_tip2 = A.commit_stop
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertHistory(
            [
                {'name': 'project-test1', 'result': 'ABORTED',
                 'changes': '1,%s' % pr_tip1},
                {'name': 'project-test2', 'result': 'ABORTED',
                 'changes': '1,%s' % pr_tip1},
                {'name': 'project-test1', 'changes': '1,%s' % pr_tip2},
                {'name': 'project-test2', 'changes': '1,%s' % pr_tip2}
            ], ordered=False
        )

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_pull_request_commented(self):

        A = self.fake_pagure.openFakePullRequest('org/project', 'master', 'A')
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent('I like that change'))
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent('recheck'))
        self.waitUntilSettled()
        self.assertEqual(4, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestInitialCommentEvent('Initial comment edited'))
        self.waitUntilSettled()
        self.assertEqual(6, len(self.history))

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_pull_request_with_dyn_reconf(self):

        zuul_yaml = [
            {'job': {
                'name': 'project-test3',
                'run': 'job.yaml'
            }},
            {'project': {
                'check': {
                    'jobs': [
                        'project-test3'
                    ]
                }
            }}
        ]
        playbook = "- hosts: all\n  tasks: []"

        A = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'A')
        A.addCommit(
            {'.zuul.yaml': yaml.dump(zuul_yaml),
             'job.yaml': playbook}
        )
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test3').result)

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_ref_updated(self):

        event = self.fake_pagure.getGitReceiveEvent('org/project')
        expected_newrev = event[1]['msg']['end_commit']
        expected_oldrev = event[1]['msg']['old_commit']
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))
        self.assertEqual(
            'SUCCESS',
            self.getJobFromHistory('project-post-job').result)

        job = self.getJobFromHistory('project-post-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual('refs/heads/master', zuulvars['ref'])
        self.assertEqual('post', zuulvars['pipeline'])
        self.assertEqual('project-post-job', zuulvars['job'])
        self.assertEqual('master', zuulvars['branch'])
        self.assertEqual(
            'https://pagure/org/project/c/%s' % zuulvars['newrev'],
            zuulvars['change_url'])
        self.assertEqual(expected_newrev, zuulvars['newrev'])
        self.assertEqual(expected_oldrev, zuulvars['oldrev'])
        self.assertEqual(expected_newrev, zuulvars['commit_id'])

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_ref_created(self):

        self.create_branch('org/project', 'stable-1.0')
        path = os.path.join(self.upstream_root, 'org/project')
        repo = git.Repo(path)
        newrev = repo.commit('refs/heads/stable-1.0').hexsha
        event = self.fake_pagure.getGitBranchEvent(
            'org/project', 'stable-1.0', 'creation', newrev)
        old = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()
        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        # New timestamp should be greater than the old timestamp
        self.assertLess(old, new)
        self.assertEqual(1, len(self.history))
        self.assertEqual(
            'SUCCESS',
            self.getJobFromHistory('project-post-job').result)
        job = self.getJobFromHistory('project-post-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual('refs/heads/stable-1.0', zuulvars['ref'])
        self.assertEqual('post', zuulvars['pipeline'])
        self.assertEqual('project-post-job', zuulvars['job'])
        self.assertEqual('stable-1.0', zuulvars['branch'])
        self.assertEqual(newrev, zuulvars['newrev'])

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_ref_deleted(self):

        event = self.fake_pagure.getGitBranchEvent(
            'org/project', 'stable-1.0', 'deletion', '0' * 40)
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_ref_updated_and_tenant_reconfigure(self):

        self.waitUntilSettled()
        old = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)

        zuul_yaml = [
            {'job': {
                'name': 'project-post-job2',
                'run': 'job.yaml'
            }},
            {'project': {
                'post': {
                    'jobs': [
                        'project-post-job2'
                    ]
                }
            }}
        ]
        playbook = "- hosts: all\n  tasks: []"
        self.create_commit(
            'org/project',
            {'.zuul.yaml': yaml.dump(zuul_yaml),
             'job.yaml': playbook},
            message='Add InRepo configuration'
        )
        event = self.fake_pagure.getGitReceiveEvent('org/project')
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()

        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        # New timestamp should be greater than the old timestamp
        self.assertLess(old, new)

        self.assertHistory(
            [{'name': 'project-post-job'},
             {'name': 'project-post-job2'},
            ], ordered=False
        )

    @simple_layout('layouts/files-pagure.yaml', driver='pagure')
    def test_pull_matched_file_event(self):
        A = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'A',
            files={'random.txt': 'test', 'build-requires': 'test'})
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        B = self.fake_pagure.openFakePullRequest('org/project', 'master', 'B',
                                                 files={'random.txt': 'test2'})
        self.fake_pagure.emitEvent(B.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        C = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'C',
            files={'build-requires': 'test'})
        self.fake_pagure.emitEvent(C.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_tag_created(self):

        path = os.path.join(self.upstream_root, 'org/project')
        repo = git.Repo(path)
        repo.create_tag('1.0')
        tagsha = repo.tags['1.0'].commit.hexsha
        event = self.fake_pagure.getGitTagCreatedEvent(
            'org/project', '1.0', tagsha)
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))
        self.assertEqual(
            'SUCCESS',
            self.getJobFromHistory('project-tag-job').result)
        job = self.getJobFromHistory('project-tag-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual('refs/tags/1.0', zuulvars['ref'])
        self.assertEqual('tag', zuulvars['pipeline'])
        self.assertEqual('project-tag-job', zuulvars['job'])
        self.assertEqual(tagsha, zuulvars['newrev'])

    @simple_layout('layouts/requirements-pagure.yaml', driver='pagure')
    def test_pr_score_require_1_vote(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project1', 'master', 'A')
        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent("I like that change"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        self.assertEqual(
            'SUCCESS',
            self.getJobFromHistory('project-test').result)

    @simple_layout('layouts/requirements-pagure.yaml', driver='pagure')
    def test_pr_score_require_2_votes(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project2', 'master', 'A')
        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent("I like that change"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsdown:"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-pagure.yaml', driver='pagure')
    def test_status_trigger(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project3', 'master', 'A')

        self.fake_pagure.emitEvent(
            A.getPullRequestStatusSetEvent("failure"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestStatusSetEvent("success"))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-pagure.yaml', driver='pagure')
    def test_tag_trigger(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project4', 'master', 'A')

        self.fake_pagure.emitEvent(
            A.getPullRequestTagAddedEvent(["lambda"]))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestTagAddedEvent(["gateit", "lambda"]))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        self.fake_pagure.emitEvent(
            A.getPullRequestTagAddedEvent(["mergeit"]))
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/requirements-pagure.yaml', driver='pagure')
    def test_tag_require(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project5', 'master', 'A')

        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        A.tags = ["lambda"]
        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        A.tags = ["lambda", "gateit"]
        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        A.tags = []
        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-pagure.yaml', driver='pagure')
    def test_flag_require(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project7', 'master', 'A')

        # CI status from other CIs must not be handled
        self.fake_pagure.emitEvent(
            A.getPullRequestStatusSetEvent("success", username="notzuul"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))
        self.assertEqual(1, len(A.flags))

        self.fake_pagure.emitEvent(
            A.getPullRequestStatusSetEvent("failure"))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))
        self.assertEqual(2, len(A.flags))

        self.fake_pagure.emitEvent(
            A.getPullRequestStatusSetEvent("success"))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))
        self.assertEqual(2, len(A.flags))

    @simple_layout('layouts/requirements-pagure.yaml', driver='pagure')
    def test_pull_request_closed(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project6', 'master', 'A')

        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # Validate a closed but not merged PR does not trigger the pipeline
        self.fake_pagure.emitEvent(A.getPullRequestClosedEvent(merged=False))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # Reset the status to Open
        # Validate a closed and merged PR triggers the pipeline
        A.status = 'Open'
        A.is_merged = False
        self.fake_pagure.emitEvent(A.getPullRequestClosedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/merging-pagure.yaml', driver='pagure')
    def test_merge_action_in_independent(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project1', 'master', 'A')
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertEqual(1, len(self.history))
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test').result)
        self.assertEqual('Merged', A.status)

    @simple_layout('layouts/merging-pagure.yaml', driver='pagure')
    def test_merge_action_in_dependent(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project2', 'master', 'A')
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        # connection.canMerge is not validated
        self.assertEqual(0, len(self.history))

        # Set the mergeable PR flag to a expected value
        A.cached_merge_status = 'MERGE'
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        # connection.canMerge is not validated
        self.assertEqual(0, len(self.history))

        # Set the score threshold as reached
        # Here we use None that means no specific score is required
        A.threshold_reached = None
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        # connection.canMerge is not validated
        self.assertEqual(0, len(self.history))

        # Set CI flag as passed CI
        A.addFlag('success', 'https://url', 'Build passed')
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        # connection.canMerge is validated
        self.assertEqual(1, len(self.history))

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test').result)
        self.assertEqual('Merged', A.status)

    @simple_layout('layouts/crd-pagure.yaml', driver='pagure')
    def test_crd_independent(self):

        # Create a change in project1 that a project2 change will depend on
        A = self.fake_pagure.openFakePullRequest('org/project1', 'master', 'A')

        # Create a commit in B that sets the dependency on A
        msg = "Depends-On: %s" % A.url
        B = self.fake_pagure.openFakePullRequest(
            'org/project2', 'master', 'B', initial_comment=msg)

        # Make an event to re-use
        event = B.getPullRequestCommentedEvent('A comment')
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()

        # The changes for the job from project2 should include the project1
        # PR content
        changes = self.getJobFromHistory(
            'project2-test', 'org/project2').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.commit_stop,
                                                   B.number,
                                                   B.commit_stop))

        # There should be no more changes in the queue
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(len(
            tenant.layout.pipeline_managers['check'].state.queues), 0)

    @simple_layout('layouts/crd-pagure.yaml', driver='pagure')
    def test_crd_dependent(self):

        # Create a change in project3 that a project4 change will depend on
        A = self.fake_pagure.openFakePullRequest('org/project3', 'master', 'A')

        # Create a commit in B that sets the dependency on A
        msg = "Depends-On: %s" % A.url
        B = self.fake_pagure.openFakePullRequest(
            'org/project4', 'master', 'B', initial_comment=msg)

        # Make an event to re-use
        event = B.getPullRequestCommentedEvent('A comment')

        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()

        # Neither A and B can't merge (no flag, no score threshold)
        self.assertEqual(0, len(self.history))

        B.threshold_reached = True
        B.addFlag('success', 'https://url', 'Build passed')
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()

        # B can't merge as A got no flag, no score threshold
        self.assertEqual(0, len(self.history))

        A.threshold_reached = True
        A.addFlag('success', 'https://url', 'Build passed')
        self.fake_pagure.emitEvent(event)
        self.waitUntilSettled()

        # The changes for the job from project4 should include the project3
        # PR content
        changes = self.getJobFromHistory(
            'project4-test', 'org/project4').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.commit_stop,
                                                   B.number,
                                                   B.commit_stop))

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)

    @simple_layout('layouts/crd-pagure.yaml', driver='pagure')
    def test_crd_needed_changes(self):

        # Given change A and B, where B depends on A, when A
        # completes B should be enqueued (using a shared queue)

        # Create a change in project3 that a project4 change will depend on
        A = self.fake_pagure.openFakePullRequest('org/project3', 'master', 'A')
        A.threshold_reached = True
        A.addFlag('success', 'https://url', 'Build passed')

        # Set B to depend on A
        msg = "Depends-On: %s" % A.url
        B = self.fake_pagure.openFakePullRequest(
            'org/project4', 'master', 'B', initial_comment=msg)
        # Make the driver aware of change B by sending an event
        # At that moment B can't merge
        self.fake_pagure.emitEvent(B.getPullRequestCommentedEvent('A comment'))
        # Now set B mergeable
        B.threshold_reached = True
        B.addFlag('success', 'https://url', 'Build passed')

        # Enqueue A, which will make the scheduler detect that B is
        # depending on so B will be enqueue as well.
        self.fake_pagure.emitEvent(A.getPullRequestCommentedEvent('A comment'))
        self.waitUntilSettled()

        # The changes for the job from project4 should include the project3
        # PR content
        changes = self.getJobFromHistory(
            'project4-test', 'org/project4').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.commit_stop,
                                                   B.number,
                                                   B.commit_stop))

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)

    @simple_layout('layouts/files-pagure.yaml', driver='pagure')
    def test_changed_file_match_filter(self):
        files = {'{:03d}.txt'.format(n): 'test' for n in range(300)}
        files["foobar-requires"] = "test"
        files["to-be-removed"] = "test"
        A = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'A', files=files)

        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        # project-test1 and project-test2 should be run
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/files-pagure.yaml', driver='pagure')
    def test_changed_and_reverted_file_not_match_filter(self):
        files = {'{:03d}.txt'.format(n): 'test' for n in range(3)}
        files["foobar-requires"] = "test"
        files["to-be-removed"] = "test"
        A = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'A', files=files)
        A.addCommit(delete_files=['to-be-removed'])

        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()
        # Only project-test1 should be run, because the file to-be-removed
        # is reverted and not in changed files to trigger project-test2
        self.assertEqual(1, len(self.history))

    @simple_layout("layouts/basic-pagure.yaml", driver="pagure")
    def test_get_project_branch_sha(self):
        # Exercise this method since it's only called from timer
        # triggers
        source = self.fake_pagure.source
        project = source.getProject('org/project')
        self.assertIsNotNone(source.getProjectBranchSha(project, 'master'))

    @simple_layout('layouts/pagure-pipeline-trigger-debug.yaml',
                   driver='pagure')
    def test_pagure_pipeline_trigger_debug(self):
        # Test that we get debug info for a pipeline debug trigger
        A = self.fake_pagure.openFakePullRequest("org/project", "master", "A")
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(1, len(A.comments))
        self.assertIn('Debug information:',
                      A.comments[0]['comment'])

        # Test the canMerge code path
        B = self.fake_pagure.openFakePullRequest("org/project", "master", "B")
        B.draft = True
        self.fake_pagure.emitEvent(
            B.getPullRequestTagAddedEvent(["gateit"]))
        self.waitUntilSettled()

        self.assertFalse(B.is_merged)
        self.assertEqual(2, len(B.comments))
        self.assertIn("can not be merged",
                      B.comments[1]['comment'])


class TestPagureToGerritCRD(ZuulTestCase):
    config_file = 'zuul-crd-pagure.conf'
    tenant_config_file = 'config/cross-source-pagure/gerrit.yaml'

    def test_crd_gate(self):
        "Test cross-repo dependencies"
        A = self.fake_pagure.openFakePullRequest('pagure/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'B')

        # A Depends-On: B
        A.editInitialComment('Depends-On: %s\n' % (B.data['url']))

        A.addFlag('success', 'https://url', 'Build passed')
        A.threshold_reached = True

        B.addApproval('Code-Review', 2)

        # Make A enter the pipeline
        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()

        # Expect not merged as B not approved yet
        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')

        for connection in self.scheds.first.connections.connections.values():
            connection.maintainCache([], max_age=0)

        B.addApproval('Approved', 1)
        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()

        self.assertTrue(A.is_merged)
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(len(A.comments), 4)
        self.assertEqual(B.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'pagure/project2').changes
        self.assertEqual(changes, '1,1 1,%s' % A.commit_stop)

    def test_crd_check(self):
        "Test cross-repo dependencies in independent pipelines"
        A = self.fake_pagure.openFakePullRequest('pagure/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange(
            'gerrit/project1', 'master', 'B')

        # A Depends-On: B
        A.editInitialComment('Depends-On: %s\n' % (B.data['url'],))

        self.executor_server.hold_jobs_in_build = True

        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        self.waitUntilSettled()

        self.assertTrue(self.builds[0].hasChanges(A, B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(len(A.comments), 2)
        self.assertEqual(B.reported, 0)

        changes = self.getJobFromHistory(
            'project-merge', 'pagure/project2').changes
        self.assertEqual(changes, '1,1 1,%s' % A.commit_stop)


class TestGerritToPagureCRD(ZuulTestCase):
    config_file = 'zuul-crd-pagure.conf'
    tenant_config_file = 'config/cross-source-pagure/gerrit.yaml'

    def test_crd_gate(self):
        "Test cross-repo dependencies"
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_pagure.openFakePullRequest('pagure/project2', 'master',
                                                 'B')

        A.addApproval('Code-Review', 2)

        AM2 = self.fake_gerrit.addFakeChange('gerrit/project1', 'master',
                                             'AM2')
        AM1 = self.fake_gerrit.addFakeChange('gerrit/project1', 'master',
                                             'AM1')
        AM2.setMerged()
        AM1.setMerged()

        # A -> AM1 -> AM2
        # A Depends-On: B
        # M2 is here to make sure it is never queried.  If it is, it
        # means zuul is walking down the entire history of merged
        # changes.

        A.setDependsOn(AM1, 1)
        AM1.setDependsOn(AM2, 1)

        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)

        for connection in self.scheds.first.connections.connections.values():
            connection.maintainCache([], max_age=0)

        B.addFlag('success', 'https://url', 'Build passed')
        B.threshold_reached = True
        self.fake_pagure.emitEvent(
            B.getPullRequestCommentedEvent(":thumbsup:"))
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))

        self.waitUntilSettled()

        self.assertEqual(AM2.queried, 0)
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertTrue(B.is_merged)
        self.assertEqual(A.reported, 2)
        self.assertEqual(len(B.comments), 3)

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,%s 1,1' % B.commit_stop)

    def test_crd_check(self):
        "Test cross-repo dependencies in independent pipelines"
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_pagure.openFakePullRequest(
            'pagure/project2', 'master', 'B')

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.executor_server.hold_jobs_in_build = True

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertTrue(self.builds[0].hasChanges(A, B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)
        self.assertEqual(A.reported, 1)
        self.assertEqual(len(B.comments), 0)

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,%s 1,1' % B.commit_stop)


class TestPagureToGithubCRD(ZuulTestCase):
    config_file = 'zuul-crd-pagure.conf'
    tenant_config_file = 'config/cross-source-pagure/github.yaml'

    def test_crd_gate(self):
        "Test cross-repo dependencies"
        A = self.fake_pagure.openFakePullRequest('pagure/project2', 'master',
                                                 'A')
        B = self.fake_github.openFakePullRequest('github/project1', 'master',
                                                 'B')
        # A Depends-On: B
        A.editInitialComment('Depends-On: %s\n' % (B.url))

        A.addFlag('success', 'https://url', 'Build passed')
        A.threshold_reached = True

        # Make A enter the pipeline
        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()

        # Expect not merged as B not approved yet
        self.assertFalse(A.is_merged)
        self.assertFalse(B.is_merged)

        for connection in self.scheds.first.connections.connections.values():
            connection.maintainCache([], max_age=0)

        B.addLabel('approved')
        self.fake_pagure.emitEvent(
            A.getPullRequestCommentedEvent(":thumbsup:"))
        self.waitUntilSettled()

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)
        self.assertEqual(len(A.comments), 4)
        self.assertEqual(len(B.comments), 2)

        changes = self.getJobFromHistory(
            'project-merge', 'pagure/project2').changes
        self.assertEqual(changes, '1,%s 1,%s' % (B.head_sha, A.commit_stop))

    def test_crd_check(self):
        "Test cross-repo dependencies in independent pipelines"
        A = self.fake_pagure.openFakePullRequest('pagure/project2', 'master',
                                                 'A')
        B = self.fake_github.openFakePullRequest('github/project1', 'master',
                                                 'A')

        # A Depends-On: B
        A.editInitialComment('Depends-On: %s\n' % B.url)

        self.executor_server.hold_jobs_in_build = True

        self.fake_pagure.emitEvent(A.getPullRequestUpdatedEvent())
        self.waitUntilSettled()

        self.assertTrue(self.builds[0].hasChanges(A, B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertFalse(B.is_merged)
        self.assertEqual(len(A.comments), 2)
        self.assertEqual(len(A.comments), 2)

        changes = self.getJobFromHistory(
            'project-merge', 'pagure/project2').changes
        self.assertEqual(changes, '1,%s 1,%s' % (B.head_sha, A.commit_stop))


class TestGithubToPagureCRD(ZuulTestCase):
    config_file = 'zuul-crd-pagure.conf'
    tenant_config_file = 'config/cross-source-pagure/github.yaml'

    # Those tests are also using the fake github implementation which
    # means that every scheduler gets a different fake github instance.
    # Thus, assertions might fail depending on which scheduler did the
    # interaction with Github.
    scheduler_count = 1

    def test_crd_gate(self):
        "Test cross-repo dependencies"
        A = self.fake_github.openFakePullRequest('github/project1', 'master',
                                                 'A')
        B = self.fake_pagure.openFakePullRequest('pagure/project2', 'master',
                                                 'B')

        # A Depends-On: B
        A.editBody('Depends-On: %s\n' % B.url)

        event = A.addLabel('approved')
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertFalse(B.is_merged)

        for connection in self.scheds.first.connections.connections.values():
            connection.maintainCache([], max_age=0)

        B.addFlag('success', 'https://url', 'Build passed')
        B.threshold_reached = True
        self.fake_pagure.emitEvent(
            B.getPullRequestCommentedEvent(":thumbsup:"))

        self.fake_github.emitEvent(event)

        self.waitUntilSettled()

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)
        self.assertEqual(len(A.comments), 2)
        self.assertEqual(len(B.comments), 3)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project1').changes
        self.assertEqual(changes, '1,%s 1,%s' % (B.commit_stop, A.head_sha))

    def test_crd_check(self):
        "Test cross-repo dependencies in independent pipelines"
        A = self.fake_github.openFakePullRequest(
            'github/project1', 'master', 'A')
        B = self.fake_pagure.openFakePullRequest(
            'pagure/project2', 'master', 'B')

        # A Depends-On: B
        self.executor_server.hold_jobs_in_build = True
        self.fake_github.emitEvent(A.editBody('Depends-On: %s\n' % B.url))
        self.waitUntilSettled()

        self.assertTrue(self.builds[0].hasChanges(A, B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertFalse(B.is_merged)
        self.assertEqual(len(A.comments), 1)
        self.assertEqual(len(B.comments), 0)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project1').changes
        self.assertEqual(changes, '1,%s 1,%s' % (B.commit_stop, A.head_sha))


class TestPagureWebhook(ZuulTestCase):
    config_file = 'zuul-pagure-driver.conf'

    def setUp(self):
        super(TestPagureWebhook, self).setUp()
        # Start the web server
        self.web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

        host = '127.0.0.1'
        # Wait until web server is started
        while True:
            port = self.web.port
            try:
                with socket.create_connection((host, port)):
                    break
            except ConnectionRefusedError:
                pass

        self.fake_pagure.setZuulWebPort(port)

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_webhook(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'A')
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent(),
                                   use_zuulweb=True,
                                   project='org/project',
                                   wrong_token=True)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent(),
                                   use_zuulweb=True,
                                   project='org/project')
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)


class TestPagureWebhookWhitelist(ZuulTestCase):
    config_file = 'zuul-pagure-driver-whitelist.conf'

    def setUp(self):
        super(TestPagureWebhookWhitelist, self).setUp()
        # Start the web server
        self.web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

        host = '127.0.0.1'
        # Wait until web server is started
        while True:
            port = self.web.port
            try:
                with socket.create_connection((host, port)):
                    break
            except ConnectionRefusedError:
                pass

        self.fake_pagure.setZuulWebPort(port)

    @simple_layout('layouts/basic-pagure.yaml', driver='pagure')
    def test_webhook_whitelist(self):

        A = self.fake_pagure.openFakePullRequest(
            'org/project', 'master', 'A')
        self.fake_pagure.emitEvent(A.getPullRequestOpenedEvent(),
                                   use_zuulweb=True,
                                   project='org/project',
                                   wrong_token=True)
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)



# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Inspect ZK contents like zk-shell; handles compressed and sharded
# data.

import argparse
import pathlib
import cmd
import sys
import textwrap
import zlib

import kazoo.client
from kazoo.exceptions import NoNodeError


def resolve_path(path, rest):
    newpath = path / rest
    newparts = []
    for part in newpath.parts:
        if part == '.':
            continue
        elif part == '..':
            newparts.pop()
        else:
            newparts.append(part)
    return pathlib.PurePosixPath(*newparts)


class REPL(cmd.Cmd):
    def __init__(self, args):
        self.path = pathlib.PurePosixPath('/')
        super().__init__()
        kwargs = {}
        if args.cert:
            kwargs['use_ssl'] = True
            kwargs['keyfile'] = args.key
            kwargs['certfile'] = args.cert
            kwargs['ca'] = args.ca
        self.client = kazoo.client.KazooClient(args.host, **kwargs)
        self.client.start()

    @property
    def prompt(self):
        return f'{self.path}> '

    def do_EOF(self, path):
        sys.exit(0)

    def do_ls(self, path):
        'List znodes: ls [PATH]'
        if path:
            mypath = self.path / path
        else:
            mypath = self.path
        try:
            for child in self.client.get_children(str(mypath)):
                print(child)
        except NoNodeError:
            print(f'No such node: {mypath}')

    def do_cd(self, path):
        'Change the working path: cd PATH'
        if path:
            newpath = resolve_path(self.path, path)
            if self.client.exists(str(newpath)):
                self.path = newpath
            else:
                print(f'No such node: {newpath}')

    def do_pwd(self):
        'Print the working path'
        print(self.path)

    def help_get(self):
        print(textwrap.dedent(self.do_get.__doc__))

    def do_get(self, args):
        """\
        Get znode value: get PATH [-v]

        -v: output metadata about the path
        """
        args = args.split(' ')
        path = args[0]
        args = args[1:]
        path = resolve_path(self.path, path)
        try:
            compressed_data, zstat = self.client.get(str(path))
        except NoNodeError:
            print(f'No such node: {path}')
            return
        was_compressed = False
        try:
            data = zlib.decompress(compressed_data)
            was_compressed = True
        except zlib.error:
            data = compressed_data
        if '-v' in args:
            print(f'Compressed: {was_compressed}')
            print(f'Size: {len(data)}')
            print(f'Compressed size: {len(compressed_data)}')
            print(f'Zstat: {zstat}')
        print(data)

    def help_unshard(self):
        print(textwrap.dedent(self.do_unshard.__doc__))

    def do_unshard(self, args):
        """\
        Get the unsharded value: get PATH [-v]

        -v: output metadata about the path
        """
        args = args.split(' ')
        path = args[0]
        args = args[1:]
        path = resolve_path(self.path, path)

        try:
            shards = sorted(self.client.get_children(str(path)))
        except NoNodeError:
            print(f'No such node: {path}')
            return
        compressed_data = b''
        data = b''
        for shard in shards:
            d, _ = self.client.get(str(path / shard))
            compressed_data += d
        if compressed_data:
            data = zlib.decompress(compressed_data)

        if '-v' in args:
            print(f'Size: {len(data)}')
            print(f'Compressed size: {len(compressed_data)}')
        print(data)

    def do_rm(self, args):
        'Delete znode: rm PATH [-r]'
        args = args.split(' ')
        path = args[0]
        args = args[1:]
        path = resolve_path(self.path, path)
        if '-r' in args:
            recursive = True
        else:
            recursive = False
        try:
            self.client.delete(str(path), recursive=recursive)
        except NoNodeError:
            print(f'No such node: {path}')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('host', help='ZK host string')
    parser.add_argument('--cert', help='Path to TLS certificate')
    parser.add_argument('--key', help='Path to TLS key')
    parser.add_argument('--ca', help='Path to TLS CA cert')
    args = parser.parse_args()

    repl = REPL(args)
    repl.cmdloop()


if __name__ == '__main__':
    main()



# Copyright 2022 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_build_event_table

Revision ID: 0ed5def089e2
Revises: c7467b642498
Create Date: 2022-12-12 12:08:20.882790

"""

# revision identifiers, used by Alembic.
revision = '0ed5def089e2'
down_revision = 'c7467b642498'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

BUILD_EVENT_TABLE = "zuul_build_event"
BUILD_TABLE = "zuul_build"


def upgrade(table_prefix=''):
    op.create_table(
        table_prefix + BUILD_EVENT_TABLE,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("build_id", sa.Integer,
                  sa.ForeignKey(table_prefix + BUILD_TABLE + ".id")),
        sa.Column("event_time", sa.DateTime),
        sa.Column("event_type", sa.String(255)),
        sa.Column("description", sa.TEXT()),
    )


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2021 BMW Group
# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
from urllib.parse import quote_plus, unquote

from kazoo.exceptions import BadVersionError, NoNodeError

from zuul.lib.logutil import get_annotated_logger
from zuul.model import SemaphoreReleaseEvent
from zuul.zk import ZooKeeperSimpleBase
from zuul.zk.event_queues import PipelineResultEventQueue


def holdersFromData(data):
    if not data:
        return []
    return json.loads(data.decode("utf8"))


def holdersToData(holders):
    return json.dumps(holders, sort_keys=True).encode("utf8")


class SemaphoreHandler(ZooKeeperSimpleBase):
    log = logging.getLogger("zuul.zk.SemaphoreHandler")

    semaphore_root = "/zuul/semaphores"
    global_semaphore_root = "/zuul/global-semaphores"

    def __init__(self, client, statsd, tenant_name, layout, abide,
                 read_only=False):
        super().__init__(client)
        if read_only:
            statsd = None
        self.read_only = read_only
        self.abide = abide
        self.layout = layout
        self.statsd = statsd
        self.tenant_name = tenant_name
        self.tenant_root = f"{self.semaphore_root}/{tenant_name}"

    def _makePath(self, semaphore):
        semaphore_key = quote_plus(semaphore.name)
        if semaphore.global_scope:
            return f"{self.global_semaphore_root}/{semaphore_key}"
        else:
            return f"{self.tenant_root}/{semaphore_key}"

    def _emitStats(self, semaphore_path, num_holders):
        if self.statsd is None:
            return
        try:
            semaphore_quoted = semaphore_path.split('/')[-1]
            semaphore_name = unquote(semaphore_quoted)
            # statsd safe key:
            semaphore_key = semaphore_name.replace('.', '_').replace('/', '_')
            key = (f'zuul.tenant.{self.tenant_name}'
                   f'.semaphore.{semaphore_key}')
            self.statsd.gauge(f'{key}.holders', num_holders)
        except Exception:
            self.log.exception("Unable to send semaphore stats:")

    def getSemaphoreInfo(self, job_semaphore):
        semaphore = self.layout.getSemaphore(self.abide, job_semaphore.name)
        return {
            'name': job_semaphore.name,
            'path': self._makePath(semaphore),
            'resources_first': job_semaphore.resources_first,
            'max': 1 if semaphore is None else semaphore.max,
        }

    def getSemaphoreHandle(self, item, job):
        return {
            "buildset_path": item.current_build_set.getPath(),
            "job_name": job.name,
        }

    def acquire(self, item, job, request_resources):
        # This is the typical method for acquiring semaphores.  It
        # runs on the scheduler and acquires all semaphores for a job.
        if self.read_only:
            raise RuntimeError("Read-only semaphore handler")
        if not job.semaphores:
            return True

        log = get_annotated_logger(self.log, item.event)
        handle = self.getSemaphoreHandle(item, job)
        infos = [self.getSemaphoreInfo(job_semaphore)
                 for job_semaphore in job.semaphores]

        return self.acquireFromInfo(log, infos, handle, request_resources)

    def acquireFromInfo(self, log, infos, handle, request_resources=False):
        # This method is used by the executor to acquire a playbook
        # semaphore; it is similar to the acquire method but the
        # semaphore info is frozen (this operates without an abide).
        if self.read_only:
            raise RuntimeError("Read-only semaphore handler")
        if not infos:
            return True

        all_acquired = True
        for info in infos:
            if not self._acquire_one(log, info, handle, request_resources):
                all_acquired = False
                break
        if not all_acquired:
            # Since we know we have less than all the required
            # semaphores, set quiet=True so we don't log an inability
            # to release them.
            self.releaseFromInfo(log, None, infos, handle, quiet=True)
            return False
        return True

    def _acquire_one(self, log, info, handle, request_resources):
        if info['resources_first'] and request_resources:
            # We're currently in the resource request phase and want to get the
            # resources before locking. So we don't need to do anything here.
            return True
        else:
            # As a safety net we want to acuire the semaphore at least in the
            # run phase so don't filter this here as re-acuiring the semaphore
            # is not a problem here if it has been already acquired before in
            # the resources phase.
            pass

        self.kazoo_client.ensure_path(info['path'])
        semaphore_holders, zstat = self.getHolders(info['path'])

        if handle in semaphore_holders:
            return True

        # semaphore is there, check max
        while len(semaphore_holders) < info['max']:
            semaphore_holders.append(handle)

            try:
                self.kazoo_client.set(info['path'],
                                      holdersToData(semaphore_holders),
                                      version=zstat.version)
            except BadVersionError:
                log.debug(
                    "Retrying semaphore %s acquire due to concurrent update",
                    info['name'])
                semaphore_holders, zstat = self.getHolders(info['path'])
                continue

            log.info("Semaphore %s acquired: handle %s",
                     info['name'], handle)
            self._emitStats(info['path'], len(semaphore_holders))
            return True

        return False

    def release(self, event_queue, item, job, quiet=False):
        if self.read_only:
            raise RuntimeError("Read-only semaphore handler")
        if not job.semaphores:
            return

        log = get_annotated_logger(self.log, item.event)

        handle = self.getSemaphoreHandle(item, job)
        infos = [self.getSemaphoreInfo(job_semaphore)
                 for job_semaphore in job.semaphores]

        return self.releaseFromInfo(log, event_queue, infos, handle,
                                    quiet=False)

    def releaseFromInfo(self, log, event_queue, infos, handle, quiet=False):
        for info in infos:
            self._release_one(log, info, handle, quiet)
            if event_queue is not None:
                # If a scheduler has been provided (which it is except
                # in the case of a rollback from acquire in this
                # class), broadcast an event to trigger pipeline runs.
                event = SemaphoreReleaseEvent(info['name'])
                if isinstance(event_queue, PipelineResultEventQueue):
                    event_queue.put(event)
                else:
                    event_queue.put(event, needs_result=False)

    def _release_one(self, log, info, handle, quiet=False):
        while True:
            try:
                semaphore_holders, zstat = self.getHolders(info['path'])
                semaphore_holders.remove(handle)
            except (ValueError, NoNodeError):
                if not quiet:
                    log.error("Semaphore %s can not be released for %s "
                              "because the semaphore is not held",
                              info['path'], handle)
                break

            try:
                self.kazoo_client.set(info['path'],
                                      holdersToData(semaphore_holders),
                                      zstat.version)
            except BadVersionError:
                log.debug(
                    "Retrying semaphore %s release due to concurrent update",
                    info['path'])
                continue

            log.info("Semaphore %s released for %s",
                     info['path'], handle)
            self._emitStats(info['path'], len(semaphore_holders))
            break

    def getHolders(self, semaphore_path):
        data, zstat = self.kazoo_client.get(semaphore_path)
        return holdersFromData(data), zstat

    def getSemaphores(self):
        ret = []
        for root in (self.global_semaphore_root, self.tenant_root):
            try:
                ret.extend(self.kazoo_client.get_children(root))
            except NoNodeError:
                pass
        return ret

    def semaphoreHolders(self, semaphore_name):
        semaphore = self.layout.getSemaphore(self.abide, semaphore_name)
        semaphore_path = self._makePath(semaphore)
        try:
            holders, _ = self.getHolders(semaphore_path)
        except NoNodeError:
            holders = []
        return holders

    def cleanupLeaks(self):
        if self.read_only:
            raise RuntimeError("Read-only semaphore handler")

        for semaphore_name in self.getSemaphores():
            for holder in self.semaphoreHolders(semaphore_name):
                if (self.kazoo_client.exists(holder["buildset_path"])
                        is not None):
                    continue

                semaphore = self.layout.getSemaphore(
                    self.abide, semaphore_name)
                info = {
                    'name': semaphore.name,
                    'path': self._makePath(semaphore),
                }
                self.log.error("Releasing leaked semaphore %s held by %s",
                               info['path'], holder)
                self._release_one(self.log, info, holder, quiet=False)



# Copyright 2024 BMW Group
# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul import model
from zuul.provider import statemachine


class StaticProviderNode(model.ProviderNode, subclass_id="static"):
    pass


class StaticInstance(statemachine.Instance):
    def __init__(self, node_config, quota):
        super().__init__()
        self.quota = quota
        self.metadata = {}
        self.private_ipv4 = None
        self.private_ipv6 = None
        self.public_ipv4 = None
        self.public_ipv6 = None
        self.cloud = None
        self.interface_ip = node_config.name

    @property
    def external_id(self):
        return self.interface_ip

    def getQuotaInformation(self):
        return self.quota
